/******************************************************************************

  TCPINFO - display configuration info to the screen

  Copyright (C) 1991, University of Waterloo
  portions Copyright (C) 1990, National Center for Supercomputer Applications

  This program is free software; you can redistribute it and/or modify
  it, but you may not sell it.

  This program is distributed in the hope that it will be useful,
  but without any warranty; without even the implied warranty of
  merchantability or fitness for a particular purpose.

      Erick Engelke                   or via E-Mail
      Faculty of Engineering
      University of Waterloo          Erick@development.watstar.uwaterloo.ca
      200 University Ave.,
      Waterloo, Ont., Canada
      N2L 3G1

  Numerous additions by G. Vanem <giva@bgnett.no>
  1995-2003

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <io.h>
#include <netdb.h>

#include "../src/wattcp.h"
#include "../src/pcconfig.h"
#include "../src/pcdbug.h"
#include "../src/pcbsd.h"
#include "../src/pctcp.h"
#include "../src/pcarp.h"
#include "../src/pcsed.h"
#include "../src/pcpkt.h"
#include "../src/pcbootp.h"
#include "../src/pcdhcp.h"
#include "../src/pcdbug.h"
#include "../src/bsdname.h"
#include "../src/udp_dom.h"
#include "../src/sock_ini.h"
#include "../src/printk.h"
#include "../src/misc.h"

#if defined(USE_BSD_FUNC)
#if defined(USE_IPV6)
  static unsigned NumHosts6Entries (void);
#endif
  static unsigned NumHostsEntries (void);
  static unsigned NumServEntries (void);
  static unsigned NumProtoEntries (void);
  static unsigned NumNetEntries (void);
#endif

char buffer [512], buf2 [512];

char *EtherAddr (const void *eth_adr)
{
  static char buf[20];
  const char *eth = (const char*)eth_adr;

  sprintf (buf, "%02X:%02X:%02X:%02X:%02X:%02X",
           eth[0] & 255, eth[1] & 255, eth[2] & 255,
           eth[3] & 255, eth[4] & 255, eth[5] & 255);
  return (buf);
}

void dump_gateways (void)
{
  struct gate_entry gw_list [10];
  int    i, num_gw = _arp_list_gateways (&gw_list[0], DIM(gw_list));

  if (num_gw == 0)
  {
    puts ("NONE\n");
    return;
  }

  puts ("GATEWAY'S IP     SUBNET           SUBNET MASK");

  for (i = 0 ; i < num_gw; i++)
  {
    char gate[20], snet[20], mask[20];

    strcpy (gate, _inet_ntoa(gate,gw_list[i].gate_ip));
    strcpy (snet, _inet_ntoa(snet,gw_list[i].subnet));
    strcpy (mask, _inet_ntoa(mask,gw_list[i].mask));

    printf ("                 : %-15s  %-15s  %-15s\n", gate,
            gw_list[i].subnet ? snet : "DEFAULT",
            gw_list[i].mask   ? mask : "DEFAULT");
  }
  puts ("");
}

void dump_arp_cache (void)
{
  struct arp_entry arp_list [100];
  int    i, num_arp = _arp_list_cache (arp_list, DIM(arp_list));

  if (num_arp == 0)
  {
    puts ("                   <None>           <None>\n");
    return;
  }

  for (i = 0; i < num_arp; i++)
  {
    char type[20];

    printf ("                 : %-15s  %s  ",
            _inet_ntoa(buffer, arp_list[i].ip),
            EtherAddr ((void*)&arp_list[i].hardware));

    if (arp_list[i].flags & ARP_FIXED)
       strcpy (type, "fixed  ");
    else if (arp_list[i].flags & ARP_DYNAMIC)
       strcpy (type, "dynamic");
    else if (arp_list[i].flags & ARP_PENDING)
       strcpy (type, "pending");
    else
       sprintf (type, "0x%02X?  ", arp_list[i].flags);
    printf ("%s ", type);

    if (arp_list[i].expiry)
    {
      long timeout = arp_list[i].expiry - set_timeout(0);
      if (timeout < 0)
           puts ("< 0 ms");
      else printf ("%ld ms\n", timeout);
    }
    else
      puts ("N/A");
  }
  puts ("");
}

const char *pkt_class_name (WORD class)
{
  return (class == PD_ETHER ? "Ethernet" :
          class == PD_TOKEN ? "Token"    :
          class == PD_SLIP  ? "SLIP"     :
          class == PD_AX25  ? "AX25"     :
          class == PD_FDDI  ? "FDDI"     :
          class == PD_PPP   ? "PPP"      :
          "Unknown");
}

const char *pkt_driver_ver (void)
{
  static char buf[10];
  int ver = pkt_get_drvr_ver();

  if (ver <= -1)
     return ("Version: unknown");
  sprintf (buf, "Version: %d.%02d", ver >> 8, ver & 0xFF);
  return (buf);
}

void usage (void)
{
  printf (" Usage: tcpinfo [-d]\n"
          "   -d sets debug-mode (WATTCP.CFG \"debug_on = 1\")\n");
  exit (0);
}

void check_setup (void)
{
  const char *watt_env  = getenv ("WATTCP.CFG");
  const char *watt_root = getenv ("WATT_ROOT");

  if (!watt_root)
     puts ("\7Warning: %%WATT_ROOT is empty.\n"
           "This is okay if you're *not* planning on programming with Watt-32");
  else
  {
    char path [512];

    strcpy (path, watt_root);
    strcat (path, "\\.");
    if (access(path,0) != 0)
       printf ("\7Warning: WATT_ROOT incorrectly set: `%s'.\n", watt_root);
  }

  if (watt_root && strchr(watt_root,'/'))
     puts ("\7Warning: %%WATT_ROOT contains forward ('/') slashes.\n"
           "This may break some Makefiles.");

  if (watt_env && strchr(watt_env,'/'))
     puts ("\7Warning: %%WATTCP.CFG contains forward slashes ('/').\n"
           "This will break non-djgpp Watt-32 programs.");
}

int main (int argc, char **argv)
{
  int  i;
  char name[50];

  if (argc >= 2 && (!strncmp(argv[1],"-?",2) || !strncmp(argv[1],"/?",2)))
     usage();

  if (argc >= 2 && (!stricmp(argv[1],"-d") || !stricmp(argv[1],"/d")))
     debug_on = 2;

  puts ("Reading Waterloo TCP configuration file.");

  surviveeth   = 1;  /* needed to not exit if pkt_eth_init() fails */
  survivebootp = 1;  /* needed to not exit if BOOTP fails */
  survivedhcp  = 1;  /* ditto for DHCP/RARP */
  surviverarp  = 1;

  dbug_init();
  sock_init();

  if (debug_on >= 2)
     puts ("");

  check_setup();

  printf ("PKTDRVR name     : %s, %s\n", _pktdrvrname, pkt_driver_ver());

#if defined(USE_MULTICAST) || defined(USE_IPV6)
  printf ("        Class    : %s\n", pkt_class_name(_pktdevclass));
  printf ("        Level    : %u\n", _pktdevlevel);
  if (_pktdevlevel >= 2 && !_pktserial)
     printf ("        RX mode  : %d\n", pkt_get_rcv_mode());
#endif

  puts ("");
  if (_pktdevclass == PD_ETHER)
     printf ("Ethernet Address : %s\n", EtherAddr(&_eth_addr));

  puts ("");
  if (multihomes)
       printf ("IP Addresses     : %s - %s\n",
               _inet_ntoa (buf2,  _gethostid()),
               _inet_ntoa (buffer,_gethostid()+multihomes));
  else printf ("IP Address       : %s\n", _inet_ntoa (buffer,_gethostid()));

  printf ("Network Mask     : %s\n\n", _inet_ntoa(buffer,sin_mask));
  printf ("Gateways         : ");
  dump_gateways();

  printf ("ARP Cache        : IP Address       MAC Address        Type    Timeout\n");
  dump_arp_cache();

  printf ("Host name        : ");
  if (gethostname(name,sizeof(name)) == 0)
       puts (name);
  else puts ("<NONE>");

  printf ("Domain name      : ");
  if (getdomainname(name,sizeof(name)) == 0)
       puts (name);
  else puts ("<NONE>");

#if 0
  printf ("Cookieserver%c    : ", (last_cookie < 2) ? ' ' : 's');
  if (!last_cookie)
     puts ("NONE DEFINED");

  for (i = 0 ; i < last_cookie; i++)
  {
    if (i)
       printf ("                 : ");
    printf ("%s\n", _inet_ntoa(buffer, cookies[i]));
  }
  puts("");
#endif

  printf ("Nameserver%c      : ", (last_nameserver < 2) ? ' ' : 's');
  if (!last_nameserver)
     puts ("NONE DEFINED\n");

  for (i = 0 ; i < last_nameserver; i++)
  {
    if (i)
       printf ("                 : ");
    printf ("%s\n", _inet_ntoa(buffer, def_nameservers[i]));
  }

  if (_bootpon)
  {
    puts("");
    printf ("BOOTP            : Enabled and %s\n", _gethostid() ? "SUCCEEDED" : "FAILED");
    printf ("BOOTP Server     : %s\n",           _inet_ntoa(buffer,_bootphost));
    printf ("BOOTP Timeout    : %i seconds\n\n", _bootptimeout);
  }
#if defined(USE_DHCP)
  else if (_dhcpon)
  {
    puts("");
    printf ("DHCP             : Enabled and %s\n", _gethostid() ? "SUCCEEDED" : "FAILED");
    printf ("DHCP Server      : %s\n",           _inet_ntoa(buffer,DHCP_get_server()));
  }
#endif

  puts("");
  printf ("Max Seg Size,MSS : %u bytes\n",   mss);
  printf ("Max Transmit,MTU : %u bytes\n\n", mtu);

  printf ("TCP timers       : Sock delay %us, Inactivity %us, Keep-alive %us\n",
          sock_delay, sock_inactive, tcp_keepalive);
  printf ("                 : Open %ums, Close %ums, RST time %ums\n",
          tcp_OPEN_TO, tcp_CLOSE_TO, tcp_RST_TIME);
  printf ("                 : RTO base %ums, RTO add %ums, Retrans %ums\n\n",
          tcp_RTO_BASE, tcp_RTO_ADD, tcp_RETRAN_TIME);

  printf ("_tcp_Socket size : %u bytes\n",   (unsigned)sizeof(_tcp_Socket));
  printf ("_udp_Socket size : %u bytes\n\n", (unsigned)sizeof(_udp_Socket));

#if defined(USE_BSD_FUNC)
  printf ("HOSTS file       : ");
  if (GetHostsFile())
       printf ("%-30s %3d entries\n", GetHostsFile(), NumHostsEntries());
  else puts ("<NONE>");

#if defined(USE_IPV6)
  printf ("HOSTS6 file      : ");
  if (GetHosts6File())
       printf ("%-30s %3d entries\n", GetHosts6File(), NumHosts6Entries());
  else puts ("<NONE>");
#endif

  printf ("SERVICES file    : ");
  if (GetServFile())
       printf ("%-30s %3d entries\n", GetServFile(), NumServEntries());
  else puts ("<NONE>");

  printf ("PROTOCOL file    : ");
  if (GetProtoFile())
       printf ("%-30s %3d entries\n", GetProtoFile(), NumProtoEntries());
  else puts ("<NONE>");

  printf ("NETWORKS file    : ");
  if (GetNetFile())
       printf ("%-30s %3d entries\n\n", GetNetFile(), NumNetEntries());
  else puts ("<NONE>\n");
#endif

  printf ("Version info     : %s\n", wattcpVersion());
  printf ("Capabilities     : %s\n", wattcpCapabilities());
  return (0);
}


#if defined(USE_BSD_FUNC)
static unsigned NumHostsEntries (void)
{
  unsigned num = 0;
  while (gethostent())
    num++;
  return (num);
}

#if defined(USE_IPV6)
static unsigned NumHosts6Entries (void)
{
  unsigned num = 0;
  while (gethostent6())
    num++;
  return (num);
}
#endif

static unsigned NumServEntries (void)
{
  unsigned num = 0;
  while (getservent())
    num++;
  return (num);
}

static unsigned NumProtoEntries (void)
{
  unsigned num = 0;
  while (getprotoent())
    num++;
  return (num);
}

static unsigned NumNetEntries (void)
{
  unsigned num = 0;
  while (getnetent())
    num++;
  return (num);
}
#endif  /* USE_BSD_FUNC */

